


package Mail::SpamAssassin::Plugin::TxRep;

use strict;
use warnings;
use re 'taint';

use NetAddr::IP 4.000;                          # qw(:upper);
use Mail::SpamAssassin::Plugin;
use Mail::SpamAssassin::Plugin::Bayes;
use Mail::SpamAssassin::Util qw(untaint_var);
use Mail::SpamAssassin::Logger;

our @ISA = qw(Mail::SpamAssassin::Plugin);


sub new {                       # constructor: register the eval rule
  my ($class, $main) = @_;

  $class   = ref($class) || $class;
  my $self = $class->SUPER::new($main);
  bless($self, $class);

  $self->{main}          = $main;
  $self->{conf}          = $main->{conf};
  $self->{factor}        = $main->{conf}->{txrep_factor};
  $self->register_eval_rule("check_senders_reputation", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
  $self->set_config($main->{conf});

  # only the default conf loaded here, do nothing here requiring
  # the runtime settings
  dbg("TxRep: new object created");
  return $self;
}


sub set_config {
  my($self, $conf) = @_;
  my @cmds;


  push (@cmds, {
    setting     => 'use_txrep',
    default     => 0,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL
  });



  push (@cmds, {
    setting     => 'txrep_factor',
    default     => 0.5,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 1.0) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_factor} = $value;
    }
  });



  push (@cmds, {
    setting     => 'txrep_dilution_factor',
    default     => 0.98,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0.7 || $value > 1.0) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_dilution_factor} = $value;
    }
  });


  push (@cmds, {
    setting     => 'txrep_expiry_days',
    default     => 365,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 10000) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_expiry_days} = $value;
    }
  });



  push (@cmds, {
    setting     => 'txrep_learn_penalty',
    default     => 20,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 200) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_learn_penalty} = $value;
    }
  });



  push (@cmds, {
    setting     => 'txrep_learn_bonus',
    default     => 20,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 200) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_learn_bonus} = $value;
    }
  });



  push (@cmds, {
    setting     => 'txrep_autolearn',
    default     => 0,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 5) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_autolearn} = $value;
    }
  });



  push (@cmds, {
    setting     => 'txrep_track_messages',
    default     => 1,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL
  });



  push (@cmds, {
    setting     => 'txrep_welcomelist_out',
    aliases	=> ['txrep_whitelist_out'], # removed in 4.1
    default     => 10,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 200) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_welcomelist_out} = $value;
    }
  });



  push (@cmds, {
    setting     => 'txrep_ipv4_mask_len',
    default     => 16,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if (!defined $value || $value eq '')
            {return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;}
        elsif ($value !~ /^\d+$/ || $value < 0 || $value > 32)
            {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_ipv4_mask_len} = $value;
    }
  });



  push (@cmds, {
    setting     => 'txrep_ipv6_mask_len',
    default     => 48,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if (!defined $value || $value eq '')
            {return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;}
        elsif ($value !~ /^\d+$/ || $value < 0 || $value > 128)
            {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_ipv6_mask_len} = $value;
    }
  });



  push (@cmds, {
    setting     => 'user_awl_sql_override_username',
    default     => '',
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING
  });



  push (@cmds, {
    setting     => 'txrep_user2global_ratio',
    default     => 0,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 10) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_user2global_ratio} = $value;
    }
  });



  push (@cmds, {
    setting     => 'auto_welcomelist_distinguish_signed',
    aliases     => ['auto_whitelist_distinguish_signed'], # removed in 4.1
    default     => 1,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL
  });



  push (@cmds, {
    setting     => 'txrep_spf',
    default     => 1,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL
  });



  push (@cmds, {
    setting     => 'txrep_weight_email',
    default     => 3,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 10) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_weight_email} = $value;
    }
  });


  push (@cmds, {
    setting     => 'txrep_weight_email_ip',
    default     => 10,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 10) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_weight_email_ip} = $value;
    }
  });


  push (@cmds, {
    setting     => 'txrep_weight_domain',
    default     => 2,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 10) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_weight_domain} = $value;
    }
  });


  push (@cmds, {
    setting     => 'txrep_weight_ip',
    default     => 4,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 10) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_weight_ip} = $value;
    }
  });


  push (@cmds, {
    setting     => 'txrep_weight_helo',
    default     => 0.5,
    type        => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code        => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value < 0 || $value > 10) {return $Mail::SpamAssassin::Conf::INVALID_VALUE;}
        $self->{txrep_weight_helo} = $value;
    }
  });



  push (@cmds, {
    setting      => 'txrep_factory',
    is_admin     => 1,
    default      => 'Mail::SpamAssassin::DBBasedAddrList',
    type         => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING
  });



  push (@cmds, {
    setting      => 'auto_welcomelist_path',
    aliases      => ['auto_whitelist_path'], # removed in 4.1
    is_admin     => 1,
    default      => '__userstate__/tx-reputation',
    type         => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING,
    code         => sub {
        my ($self, $key, $value, $line) = @_;
        unless (defined $value && $value !~ /^$/) {return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;}
        $self->{auto_welcomelist_path} = $value;
    }
  });



  push (@cmds, {
    setting      => 'auto_welcomelist_db_modules',
    aliases      => ['auto_whitelist_db_modules'], # removed in 4.1
    is_admin     => 1,
    default      => 'DB_File GDBM_File SDBM_File',
    type         => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING
  });



  push (@cmds, {
    setting      => 'auto_welcomelist_file_mode',
    aliases      => ['auto_whitelist_file_mode'], # removed in 4.1
    is_admin     => 1,
    default      => '0700',
    type         => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    code         => sub {
        my ($self, $key, $value, $line) = @_;
        if ($value !~ /^0?[0-7]{3}$/) {
            return $Mail::SpamAssassin::Conf::INVALID_VALUE;
        }
        $value = '0'.$value if length($value) == 3; # Bug 5771
        $self->{auto_welcomelist_file_mode} = untaint_var($value);
    }
  });



  push (@cmds, {
    setting      => 'user_awl_dsn',
    is_admin     => 1,
    type         => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING
  });



  push (@cmds, {
    setting      => 'user_awl_sql_username',
    is_admin     => 1,
    type         => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING
  });



  push (@cmds, {
    setting      => 'user_awl_sql_password',
    is_admin     => 1,
    type         => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING
  });



  push (@cmds, {
    setting      => 'user_awl_sql_table',
    is_admin     => 1,
    default      => 'txrep',
    type         => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING
  });

  $conf->{parser}->register_commands(\@cmds);
}


sub _message {
  my ($self, $value, $msg) = @_;
  print "SpamAssassin TxRep: $value\n" if ($msg);
  dbg("TxRep: $value");
}


sub _fail_exit {
  my ($self, $err) = @_;
  my $eval_stat = ($err ne '') ? $err : "errno=$!";
  chomp $eval_stat;
  warn("TxRep: open of TxRep file failed: $eval_stat\n");
  if (!defined $self->{txKeepStoreTied}) {$self->finish();}
  return 0;
}


sub _fn_envelope {
  my ($self, $args, $value, $msg) = @_;

  unless ($self->{main}->{conf}->{use_txrep}){                                  return 0;}
  unless ($args->{address}) {$self->_message($args->{cli_p},"failed ".$msg);    return 0;}

  my $factor =	$self->{conf}->{txrep_weight_email} +
		$self->{conf}->{txrep_weight_email_ip} +
		$self->{conf}->{txrep_weight_domain} +
		$self->{conf}->{txrep_weight_ip} +
		$self->{conf}->{txrep_weight_helo};
  my $sign = $args->{signedby};
  my $id     = $args->{address};
  if (index($args->{address}, ',') >= 0) {
    $sign = $args->{address};
    $sign =~ s/^.*,//g;
    $id   =~ s/,.*$//g;
  }

  # simplified regex used for IP detection (possible FP at a domain is not critical)
  if ($id !~ /\./ && $self->{conf}->{txrep_weight_helo}) 
	{$factor /= $self->{conf}->{txrep_weight_helo}; $sign = 'helo';}
  elsif ($id =~ /^[a-f\d\.:]+$/ && $self->{conf}->{txrep_weight_ip})
	{$factor /= $self->{conf}->{txrep_weight_ip};}
  elsif (index($id, '@') >= 0 && $self->{conf}->{txrep_weight_email})
	{$factor /= $self->{conf}->{txrep_weight_email};}
  elsif (index($id, '@') == -1 && $self->{conf}->{txrep_weight_domain})
	{$factor /= $self->{conf}->{txrep_weight_domain};}
  else	{$factor  = 1;}

  $self->open_storages();
  my $score  = (!defined $value)? undef : $factor * $value;
  my $status = $self->modify_reputation($id, $score, $sign);
  dbg("TxRep: $msg %s (score %s) %s", $id, $score || 'undef', $sign || '');
  eval {
    $self->_message($args->{cli_p}, ($status?"":"error ") . $msg . ": " . $id);
    if (!defined $self->{txKeepStoreTied}) {$self->finish();}
    1;
  } or return $self->_fail_exit( $@ );
  return $status;
}



sub blocklist_address {my $self=shift; return $self->_fn_envelope(@_,  100, "blocklisting address");}
*blacklist_address = \&blocklist_address; # removed in 4.1
sub welcomelist_address {my $self=shift; return $self->_fn_envelope(@_, -100, "welcomelisting address");}
*whitelist_address = \&welcomelist_address; # removed in 4.1
sub remove_address    {my $self=shift; return $self->_fn_envelope(@_,undef, "removing address");}



sub check_senders_reputation {
  my ($self, $pms) = @_;


  my $autolearn = defined $self->{autolearn};
  $self->{last_pms} = $self->{autolearn} = undef;

  # Cases where we would not be able to use TxRep
  return 0 unless ($self->{conf}->{use_txrep});
  if ($self->{conf}->{use_auto_welcomelist}) {
    warn("TxRep: cannot run when Auto-Welcomelist is enabled. Please disable it!\n");
    return 0;
  }
  if ($autolearn && !$self->{conf}->{txrep_autolearn}) {
    dbg("TxRep: autolearning disabled, no more reputation adjusting, quitting");
    return 0;
  }
  my @from = $pms->all_from_addrs();
  if (@from && $from[0] eq 'ignore@compiling.spamassassin.taint.org') {
    dbg("TxRep: no scan in lint mode, quitting");
    return 0;
  }

  my $delta    = 0;
  my $timer    = $self->{main}->time_method("total_txrep");
  my $msgscore = (defined $self->{learning})? $self->{learning} : $pms->get_autolearn_points();
  my $date     = $pms->{msg}->receive_date() || $pms->{date_header_time};
  my $msg_id   = $self->{msgid} || $pms->{msg}->generate_msgid();

  my $from   = lc $pms->get('From:addr') || $pms->get('EnvelopeFrom:addr');
  return 0 unless $from =~ /\S/;
  my $domain = $from;
  $domain =~ s/^.+@//;

  # Find the last untrusted relay and populate helo and original IP
  my ($origip, $helo);
  if (defined $pms->{relays_trusted} || defined $pms->{relays_untrusted}) {
    my $trusteds = @{$pms->{relays_trusted}};
    foreach my $rly ( @{$pms->{relays_trusted}}, @{$pms->{relays_untrusted}} ) {
	# Get the last found HELO, regardless of private/public or trusted/untrusted
	# Avoiding a redundant duplicate entry if HELO is equal/similar to another identificator
	if (defined $rly->{helo} &&
            $rly->{helo} !~ /^\[?\Q$rly->{ip}\E\]?$/ &&
            $rly->{helo} !~ /^\Q$domain\E$/i &&
            $rly->{helo} !~ /^\Q$from\E$/i ) {
	    $helo   = $rly->{helo};
	}
	# use only trusted ID, but use the first untrusted IP (if available) (AWL bug 6908)
	# at low spam scores (<2) ignore trusted/untrusted
	# set IP to 127.0.0.1 for any internal IP, so that it can be distinguished from none (AWL bug 6357)
	if ((--$trusteds >=  0 || $msgscore<2) && !$msg_id && $rly->{id})            {$msg_id = $rly->{id};}
	if (($trusteds   >= -1 || $msgscore<2) && !$rly->{ip_private} && $rly->{ip}) {$origip = $rly->{ip};}
	if ( $trusteds   >=  0     && !$origip &&  $rly->{ip_private} && $rly->{ip}) {$origip = '127.0.0.1';}
    }
  }

  # Look for previous scores of the same message, for instance when doing re-learning
  if ($self->{conf}->{txrep_track_messages}) {
    if ($msg_id) {
        my $msg_rep = $self->check_reputations($pms, 'MSG_ID', $msg_id, undef, $date, undef);
        if (defined $msg_rep && $self->count()) {
            if (defined $self->{learning} && !defined $self->{forgetting}) {
                # already learned, forget only if already learned (count>1), and relearn
                # when only scanned (count=1), go ahead with normal rep scan
                if ($self->count() > 1) {
                    $self->{last_pms} = $pms;                   # cache the pmstatus
                    $self->forget_message($pms->{msg},$msg_id); # sub reentrance OK
                }
            } elsif ($self->{forgetting}) {
                $msgscore = $msg_rep;   # forget the old stored score instead of the one got now
                dbg("TxRep: forgetting stored score %0.3f of message %s", $msgscore || 'undef', $msg_id);
            } else {
                # calculating the delta from the stored message reputation
                $delta = ($msgscore + $self->{conf}->{txrep_factor}*$msg_rep) / (1+$self->{conf}->{txrep_factor}) - $msgscore;
                if ($delta != 0) {
                    $pms->got_hit("TXREP", "TXREP: ", ruletype => 'eval', score => sprintf("%0.3f", $delta));
                }
                dbg("TxRep: message %s already scanned, using old data; post-TxRep score: %0.3f", $msg_id, $pms->{score} || 'undef');
                return 0;
            }
        }       # no stored reputation found, go ahead with normal rep scan
    } else {dbg("TxRep: no message-id available, parsing forced");}
  }             # else no message tracking, go ahead with normal rep scan

  # welcomelists recipients at senders from internal networks after checking MSG_ID only
  if ( $self->{conf}->{txrep_welcomelist_out} &&
          defined $pms->{relays_internal} &&  @{$pms->{relays_internal}} &&
        (!defined $pms->{relays_external} || !@{$pms->{relays_external}})
     ) {
    foreach my $rcpt ($pms->all_to_addrs()) {
        if ($rcpt) {
            dbg("TxRep: internal sender, welcomelisting recipient: $rcpt");
            $self->modify_reputation($rcpt, -1*$self->{conf}->{txrep_welcomelist_out}, undef);
        }
    }
  }

  # Get the signing domain
  my $signedby = ($self->{conf}->{auto_welcomelist_distinguish_signed})? $pms->get_tag('DKIMDOMAIN') : undef;

  # Summary of all information we've gathered so far
  dbg("TxRep: active, %s pre-score: %s, autolearn score: %s, IP: %s, address: %s %s",
    $msg_id       || '',
    $pms->{score} || '?',
    $msgscore     || '?',
    $origip       || '?',
    $from         || '?',
    $signedby ? "signed by $signedby" : '(unsigned)'
  );

  my $ip = $origip;
  if ($signedby) {
    $ip       = undef;
    $domain   = $signedby;
  } elsif ($pms->{spf_pass} && $self->{conf}->{txrep_spf}) {
    $ip       = undef;
    $signedby = 'spf';
  }

  my $totalweight      = 0;
  $self->{totalweight} = $totalweight;

  # Get current reputation info
  $delta += $self->check_reputations($pms, 'EMAIL_IP', $from, $ip, $signedby, $msgscore);

  if ($domain) {
    $delta += $self->check_reputations($pms, 'DOMAIN', $domain, $ip, $signedby, $msgscore);
  }
  if ($helo) {
    $delta += $self->check_reputations($pms, 'HELO', $helo, undef, 'HELO', $msgscore);
  }
  if ($origip) {
    if (!$signedby) {
      $delta += $self->check_reputations($pms, 'EMAIL', $from, undef, undef, $msgscore);
    }
    $delta += $self->check_reputations($pms, 'IP', $origip, undef, undef, $msgscore);
  }

  # Learn against this message and store reputation
  if (!defined $self->{learning}) {
    $delta = ($self->{totalweight})? $self->{conf}->{txrep_factor} * $delta / $self->{totalweight}  :  0;
    if ($delta) {
      $pms->got_hit("TXREP", "TXREP: ", ruletype => 'eval', score => sprintf("%0.3f", $delta));
    }
    $msgscore += $delta;
    if (defined $pms->{score}) {
      dbg("TxRep: post-TxRep score: %.3f", $pms->{score});
    }
  }
  # Track message ID
  if ($self->{conf}->{txrep_track_messages} && $msg_id) {
    $self->check_reputations($pms, 'MSG_ID', $msg_id, undef, $date, $msgscore);
  }
  # Close any open resources
  if (!defined $self->{txKeepStoreTied}) {
    $self->finish();
  }

  return 0;
}


sub check_reputations {
  my $self = shift;
  my $delta;

  if ($self->open_storages()) {
    if ($self->{conf}->{txrep_user2global_ratio} && $self->{user_storage} != $self->{global_storage}) {
      my $user   = $self->check_reputation('user_storage',  @_);
      my $global = $self->check_reputation('global_storage',@_);

      if (defined $user and $user == $user) {
        $delta = ( $self->{conf}->{txrep_user2global_ratio} * $user + $global ) / ( 1 + $self->{conf}->{txrep_user2global_ratio} );
      } else {
        $delta = $global;
      }
    } else {
      $delta = $self->check_reputation(undef,@_);
    }
  }
  return $delta;
}


sub check_reputation {
  my ($self, $storage, $pms, $key, $id, $ip, $signedby, $msgscore) = @_;

  my $delta  = 0;
  my $weight = ($key eq 'MSG_ID') ? 1 : $pms->{main}->{conf}->{'txrep_weight_'.lc($key)};



  if (defined $weight && $weight) {
    my $meanrep;
    my $timer = $self->{main}->time_method('check_txrep_'.lc($key));

    if (defined $storage) {
        $self->{checker} = $self->{$storage};
    }
    my $found  = $self->get_sender($id, $ip, $signedby);
    my $tag_id = (defined $storage)? uc($key.'_'.substr($storage,0,1)) : uc($key);
    # TEMPLATE TAGS should match [A-Z] in their name
    # and "_" must be avoided
    $tag_id =~ s/_//g;
    if (defined $found && $self->count()) {
        $meanrep = $self->total() / $self->count();
    }
    if ($self->{learning} && defined $msgscore) {
        if (defined $meanrep) {
            # $msgscore<=>0 gives the sign of $msgscore
            $msgscore += ($msgscore<=>0) * abs($meanrep);
        }
        dbg("TxRep: reputation: %s, count: %d, learning: %s, $tag_id: %s",
            defined $meanrep? sprintf("%.3f",$meanrep) : 'none',
            $self->count()      || 0,
            $self->{learning}   || '',
            $id                 || 'none'
        );
    } else {
        $self->{totalweight} += $weight;
        if ($key eq 'MSG_ID' && $self->count() > 0) {
            $delta = $self->total() / $self->count();
	    $pms->set_tag('TXREP'.$tag_id,              sprintf("%2.1f", $delta));
        } elsif (defined $self->total()) {
            #Bug 7164 - $msgscore undefined
            if (defined $msgscore) {
              $delta = ($self->total() + $msgscore) / (1 + $self->count()) - $msgscore;
            } else {
              $delta = ($self->total()) / (1 + $self->count());
            }

            $pms->set_tag('TXREP'.$tag_id,             sprintf("%2.1f", $delta));
            if (defined $meanrep) {
                $pms->set_tag('TXREP'.$tag_id.'MEAN', sprintf("%2.1f", $meanrep));
            }
            $pms->set_tag('TXREP'.$tag_id.'COUNT',    sprintf("%2.1f", $self->count()));
            $pms->set_tag('TXREP'.$tag_id.'PRESCORE', sprintf("%2.1f", $pms->{score}));
        } else {
            $pms->set_tag('TXREP'.$tag_id.'UNKNOWN', 1);
        }
        dbg("TxRep: reputation: %s, count: %d, weight: %.1f, delta: %.3f, $tag_id: %s",
            defined $meanrep? sprintf("%.3f",$meanrep) : 'none',
            $self->count()      || 0,
            $weight             || 0,
            $delta              || 0,
            $id                 || 'none'
        );
    }
    $timer = $self->{main}->time_method('update_txrep_'.lc($key));
    if (defined $msgscore) {
        if ($self->{forgetting}) {              # forgetting a message score
            $self->remove_score($msgscore);     # remove the given score and decrement the count
            if ($key eq 'MSG_ID') {             # remove the message ID score completely
                $self->{checker}->remove_entry($self->{entry});
            }
        } else {
            $self->add_score($msgscore);        # add the score and increment the count
            if ($self->{learning} && $key eq 'MSG_ID' && $self->count() eq 1) {
                $self->add_score($msgscore);    # increasing the count by 1 at a learned score (count=2)
            }                                   # it can be distinguished from a scanned score (count=1)
        }
    } elsif (defined $found && $self->{forgetting} && $key eq 'MSG_ID') {
        $self->{checker}->remove_entry($self->{entry}); #forgetting the message ID
    }
  }
  if (!defined $storage) {
    $self->{checker} = $self->{default_storage};
  }

  return ($weight || 0) * ($delta || 0);
}




sub count {my $self=shift;  return (defined $self->{checker})? $self->{entry}->{msgcount}    : undef;}
sub total {my $self=shift;  return (defined $self->{checker})? $self->{entry}->{totscore} : undef;}


sub get_sender {
  my ($self, $addr, $origip, $signedby) = @_;

  return unless (defined $self->{checker});

  my $fulladdr   = $self->pack_addr($addr, $origip);
  my $entry      = $self->{checker}->get_addr_entry($fulladdr, $signedby);
  $self->{entry} = $entry;
  $origip        = $origip || 'none';

  if ($entry->{msgcount}<0 || $entry->{msgcount}=~/^(nan|)$/ || $entry->{totscore}=~/^(nan|)$/) {
    warn "TxRep: resetting bad data for ($addr, $origip), count: $entry->{msgcount}, totscore: $entry->{totscore}\n";
    $self->{entry}->{msgcount} = $self->{entry}->{totscore} = 0;
  }
  return $self->{entry}->{msgcount};
}


sub add_score {
  my ($self,$score) = @_;

  return unless (defined $self->{checker});       # no factory defined; we can't check

  if ($score != $score) {
    warn "TxRep: attempt to add a $score to TxRep entry ignored\n";
    return;                                       # don't try to add a NaN
  }
  $self->{entry}->{msgcount} ||= 0;

  # performing the dilution aging correction
  if (defined $self->total() && defined $self->count() && defined $self->{txrep_dilution_factor}) {
    my $diluted_total =
        ($self->count() + 1) *
        ($self->{txrep_dilution_factor} * $self->total() + $score) /
        ($self->{txrep_dilution_factor} * $self->count() + 1);
    my $corrected_score = $diluted_total - $self->total();
    $self->{checker}->add_score($self->{entry}, $corrected_score);
  } else {
    $self->{checker}->add_score($self->{entry}, $score);
  }
}



sub remove_score {
  my ($self,$score) = @_;

  return unless (defined $self->{checker});       # no factory defined; we can't check

  if ($score != $score) {                               # don't try to add a NaN
    warn "TxRep: attempt to add a $score to TxRep entry ignored\n";
    return;
  }
  # no reversal dilution aging correction (not easily possible),
  # just removing the original message score
  if ($self->{entry}->{msgcount} > 2)
        {$self->{entry}->{msgcount} -= 2;}
  else  {$self->{entry}->{msgcount}  = 0;}
  # subtract 2, and add a score; hence decrementing by 1
  $self->{checker}->add_score($self->{entry}, -1*$score);
}



sub modify_reputation {
  my ($self, $addr, $score, $signedby) = @_;

  return unless (defined $self->{checker});       # no factory defined; we can't check
  my $fulladdr = $self->pack_addr($addr, undef);
  my $entry    = $self->{checker}->get_addr_entry($fulladdr, $signedby);

  # remove any old entries (will remove per-ip entries as well)
  # always call this regardless, as the current entry may have 0
  # scores, but the per-ip one may have more
  $self->{checker}->remove_entry($entry);

  # remove address only, no new score to add if score NaN or undef
  if (defined $score && $score==$score) {
    # else add score. get a new entry first
    $entry = $self->{checker}->get_addr_entry($fulladdr, $signedby);
    $self->{checker}->add_score($entry, $score);
  }
  return 1;
}


sub open_storages {
  my $self = shift;

  # Enabled per bug 7191 comment 18
  return 1 unless (!defined $self->{default_storage});

  return 1 if defined ($self->{checker});

  my $factory;
  if ($self->{main}->{pers_addr_list_factory}) {
    $factory = $self->{main}->{pers_addr_list_factory};
  } else {
    my $type = $self->{conf}->{txrep_factory};
    if ($type =~ /^[_A-Za-z0-9:]+$/) {
        $type = untaint_var($type);
        eval '
          require '.$type.';
          $factory = '.$type.'->new();
          1;
        ' or do {
            my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
            warn "TxRep: $eval_stat\n";
            undef $factory;
        };
        $self->{main}->set_persistent_address_list_factory($factory) if $factory;
    } else {warn "TxRep: illegal factory setting\n";}
  }
  if (defined $factory) {
    $self->{checker} = $self->{default_storage} = $factory->new_checker($self->{main});

    if ($self->{conf}->{txrep_user2global_ratio} && !defined $self->{global_storage}) {
	# hack to handle the BDB and SQL factory types of the storage object
	# TODO: add an a method to the handler class instead
	my ($storage_type, $is_global);
	
	if (index(ref($factory), 'SQLBasedAddrList') >= 0) {
	    $is_global    = defined $self->{conf}->{user_awl_sql_override_username};
	    $storage_type = 'SQL';
	    if ($is_global && $self->{conf}->{user_awl_sql_override_username} eq $self->{main}->{username}) {
		# skip double storage if current user same as the global override
		$self->{user_storage} = $self->{global_storage} = $self->{default_storage};
	    }
	} elsif (index(ref($factory), 'DBBasedAddrList') >= 0) {
	    $is_global    = index($self->{conf}->{auto_welcomelist_path}, '__userstate__') == -1;
	    $storage_type = 'DB';
	}
	if (!defined $self->{global_storage}) {
	    my $sql_override_orig = $self->{conf}->{user_awl_sql_override_username};
	    my $awl_path_orig     = $self->{conf}->{auto_welcomelist_path};
	    if ($is_global) {
		$self->{conf}->{user_awl_sql_override_username} = '';
		$self->{conf}->{auto_welcomelist_path}            = '__userstate__/tx-reputation';
		$self->{global_storage} = $self->{default_storage};
		$self->{user_storage}   = $factory->new_checker($self->{main});
	    } else {
		$self->{conf}->{user_awl_sql_override_username} = 'GLOBAL';
		$self->{conf}->{auto_welcomelist_path}            = '__local_state_dir__/tx-reputation';
		$self->{global_storage} = $factory->new_checker($self->{main});
		$self->{user_storage}   = $self->{default_storage};
	    }
	    $self->{conf}->{user_awl_sql_override_username} = $sql_override_orig;
	    $self->{conf}->{auto_welcomelist_path}            = $awl_path_orig;
	
	    # Another ugly hack to find out whether the user differs from
	    # the global one. We need to add a method to the factory handlers
	    if ($storage_type eq 'DB' && 
		$self->{user_storage}->{locked_file} eq $self->{global_storage}->{locked_file}) {
		if ($is_global) 
		     {$self->{global_storage}->finish();}
		else {$self->{user_storage}->finish();}
		$self->{user_storage} = $self->{global_storage} = $self->{default_storage};
	    }
	}
    }
  } else {
    $self->{user_storage} = $self->{global_storage} = $self->{checker} = $self->{default_storage} = undef;
    warn("TxRep: could not open storages, quitting!\n");
    return 0;
  }
  return 1;
}


sub finish {
  my $self = shift;

  return unless (defined $self->{checker});       # no factory defined; we can't check

  if ($self->{conf}->{txrep_user2global_ratio} && defined $self->{user_storage} && ($self->{user_storage} != $self->{global_storage})) {
    $self->{user_storage}->finish();
    $self->{global_storage}->finish();
    $self->{user_storage}   = undef;
    $self->{global_storage} = undef;
  } elsif (defined $self->{default_storage}) {
    $self->{default_storage}->finish();
  }
 $self->{default_storage} = $self->{checker} = undef;
 $self->{factory} = undef;
}


sub ip_to_awl_key {
  my ($self, $origip) = @_;

  my $result;
  local $1;
  if (!defined $origip) {
    # could not find an IP address to use
  } elsif ($origip =~ /^ (\d{1,3} \. \d{1,3}) \. \d{1,3} \. \d{1,3} $/xs) {
    my $mask_len = $self->{conf}->{txrep_ipv4_mask_len};
    $mask_len = 16  if !defined $mask_len;
    # handle the default and easy cases manually
    if    ($mask_len == 32) {$result = $origip;}
    elsif ($mask_len == 16) {$result = $1;}
    else {
      my $origip_obj = NetAddr::IP->new($origip . '/' . $mask_len);
      if (!defined $origip_obj) {                       # invalid IPv4 address
        dbg("TxRep: bad IPv4 address $origip");
      } else {
        $result =        $origip_obj->network->addr;
        $result =~s/(\.0){1,3}\z//;                     # truncate zero tail
      }
    }
  } elsif (index($origip, ':') >= 0 &&                            # triage
           $origip =~
           /^ [0-9a-f]{0,4} (?: : [0-9a-f]{0,4} | \. [0-9]{1,3} ){2,9} $/xsi) {
    # looks like an IPv6 address
    my $mask_len = $self->{conf}->{txrep_ipv6_mask_len};
    $mask_len = 48  if !defined $mask_len;
    my $origip_obj = NetAddr::IP->new6($origip . '/' . $mask_len);
    if (!defined $origip_obj) {                         # invalid IPv6 address
      dbg("TxRep: bad IPv6 address $origip");
    } else {
      $result = $origip_obj->network->full6;            # string in a canonical form
      $result =~ s/(:0000){1,7}\z/::/;                  # compress zero tail
    }
  } else {
    dbg("TxRep: bad IP address $origip");
  }
  if (defined $result && length($result) > 39) {        # just in case, keep under
    $result = substr($result,0,39);                     # the awl.ip field size
  }
  return $result;
}


sub pack_addr {
  my ($self, $addr, $origip) = @_;

  $addr = lc $addr;
  $addr =~ s/[\000\;\'\"\!\|]/_/gs;                     # paranoia

  if ( defined $origip) {$origip = $self->ip_to_awl_key($origip);}
  if (!defined $origip) {$origip = 'none';}
  return $addr . "|ip=" . $origip;
}



sub learner_new {
  my ($self) = @_;

  $self->{txKeepStoreTied} = undef;
  return $self;
}


sub autolearn {
  my ($self, $params) = @_;

  $self->{last_pms} = $params->{permsgstatus};
  return $self->{autolearn} = 1;
}


sub learn_message {
  my ($self, $params) = @_;
  return 0 unless (defined $params->{isspam});

  dbg("TxRep: learning a message");
  my $pms = ($self->{last_pms})? $self->{last_pms} : Mail::SpamAssassin::PerMsgStatus->new($self->{main}, $params->{msg});
  if (!defined $pms->{relays_internal} && !defined $pms->{relays_external}) {
    $pms->extract_message_metadata();
  }

  if ($params->{isspam})
        {$self->{learning} =      $self->{conf}->{txrep_learn_penalty};}
  else  {$self->{learning} = -1 * $self->{conf}->{txrep_learn_bonus};}

  my $ret = !$self->{learning} || $self->check_senders_reputation($pms);
  $self->{learning} = undef;
  return $ret;
}


sub forget_message {
  my ($self, $params) = @_;
  return 0 unless ($self->{conf}->{use_txrep});
  my $pms = ($self->{last_pms})? $self->{last_pms} : Mail::SpamAssassin::PerMsgStatus->new($self->{main}, $params->{msg});

  dbg("TxRep: forgetting a message");
  $self->{forgetting} = 1;
  my $ret = $self->check_senders_reputation($pms);
  $self->{forgetting} = undef;
  return $ret;
}


sub learner_expire_old_training {
  my ($self, $params) = @_;
  return 0 unless ($self->{conf}->{use_txrep} && $self->{conf}->{txrep_expiry_days});

  dbg("TxRep: expiry not implemented yet");
}


sub learner_close {
  my ($self, $params) = @_;
  my $quiet = $params->{quiet};
  return 0 unless ($self->{conf}->{use_txrep});

  $self->{txKeepStoreTied} = undef;
  $self->finish();
  dbg("TxRep: learner_close");
}



1;
