
package Mail::SpamAssassin::Plugin::PhishTag;

use strict;
use warnings;
use re 'taint';
use Errno qw(EBADF);

use Mail::SpamAssassin::Plugin;
use Mail::SpamAssassin::Logger;

our @ISA = qw(Mail::SpamAssassin::Plugin);

sub new{
  my ($class, $mailsa)=@_;
  $class=ref($class) ||$class;
  my $self = $class->SUPER::new($mailsa);
  bless($self,$class);
  $self->set_config($mailsa->{conf});
  return $self;
}

sub set_config{
  my($self, $conf) = @_;
  my @cmds;

  push (@cmds, {
    setting => 'trigger_target',
    type => $Mail::SpamAssassin::Conf::CONF_TYPE_HASH_KEY_VALUE,
    is_admin => 1,
  });

  push (@cmds, {
    setting => 'trigger_config',
    type => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING,
    is_admin => 1,
    default => '',
  });

  push (@cmds, {
    setting => 'trigger_ratio',
    type => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
    is_admin => 1,
    default =>  0,
  });

  $conf->{parser}->register_commands(\@cmds);
}

sub check_start{
  my ($self, $params) = @_;
  my $pms = $params->{permsgstatus};
  
  #initialize the PHISHTAG data structure for 
  #saving configuration information
  $pms->{PHISHTAG} = {};
  $pms->{PHISHTAG}->{triggers}={};
  $pms->{PHISHTAG}->{targets}=[];

  #read the configuration info
  $self->read_configfile($params);
  $self->read_settings($params);
}

sub read_settings{
  my ($self, $params) = @_;
  my $pms = $params->{permsgstatus};

  my $triggers= $pms->{PHISHTAG}->{triggers};
  my $targets= $pms->{PHISHTAG}->{targets};
  while (my ($tname,$ttarget)=each %{$pms->{conf}->{trigger_target}}){
      push @$targets, [$ttarget, $tname];
      $$triggers{$tname}=0;
  }
}


sub read_configfile{
  my ($self, $params) = @_;
  my $pms = $params->{permsgstatus};

  #nothing interesting here if there is not a configuration file
  return if($pms->{conf}->{trigger_config} !~/\S/);

  my $triggers= $pms->{PHISHTAG}->{triggers};
  my $targets= $pms->{PHISHTAG}->{targets};

  my $target;
  local *F;
  open(F, '<', $pms->{conf}->{trigger_config});
  for ($!=0; <F>; $!=0) {
      #each entry is separated by blank lines
      undef($target) if(!/\S/);

      #lines that start with pound are comments
      next if(/^\s*\#/);

      #an entry starts with a URL line prefixed with the word "target"
      if(/^target\s+(\S+)/){
	  $target=[$1];
	  push @$targets,$target;
      }
      #add the test to the list of listened triggers
      #and          to the triggers of the last target
      elsif(defined $target){
	  s/\s+//g;
	  $$triggers{$_}=0;
	  push @$target, $_;
      }
  }
  defined $_ || $!==0  or
    $!==EBADF ? dbg("PHISHTAG: error reading config file: $!")
              : die "error reading config file: $!";
  close(F)  or die "error closing config file: $!";
}

sub hit_rule {
  my ($self, $params) = @_;
  my $pms = $params->{permsgstatus};
  my $rulename = $params->{rulename};
 
  #mark the rule as hit
  if(defined($pms->{PHISHTAG}->{triggers}->{$rulename})){
      $pms->{PHISHTAG}->{triggers}->{$rulename}=1;
      dbg("PHISHTAG: $rulename has been caught\n");
  }
}

sub check_post_learn {
  my ($self, $params) = @_;
  my $pms = $params->{permsgstatus};

  #find out which targets have fulfilled their requirements
  my $triggers= $pms->{PHISHTAG}->{triggers};
  my $targets= $pms->{PHISHTAG}->{targets};
  my @filled;
  foreach my $target(@$targets){
      my $uri= $$target[0];
      my $fulfilled=1;
      #all the triggers of a target have to exist for it to be fulfilled
      foreach my $i(1..$#$target){
	  if(! $triggers->{$$target[$i]}){
	      $fulfilled=0;
	      last;
	  }
      }
      if($fulfilled){
	  push @filled, $uri;
	  dbg("PHISHTAG: Fulfilled $uri\n");
      }
  }
  
  if(scalar(@filled) &&
     $pms->{conf}->{trigger_ratio} > rand(100)){
      $pms->{PHISHTAG}->{letgo}=0;
      $pms->{PHISHTAG}->{uri}=$filled[int(rand(scalar(@filled)))];
      
      dbg("PHISHTAG: Decided to keep this email and point to ". 
	  $pms->{PHISHTAG}->{uri});
      #make sure that SpamAssassin does not remove this email
      $pms->got_hit("PHISHTAG_TOSS", 
		    "BODY: ", 
		    score => -100);
  }
  else{
      dbg("PHISHTAG: Will let this email to SpamAssassin's discretion\n");
      $pms->{PHISHTAG}->{letgo}=1;
  }
  

  #nothing interesting here, if we will not rewrite the email
  if($pms->{PHISHTAG}->{letgo}){
      return;
  }
  
  my $pristine_body=\$pms->{msg}->{pristine_body};
  #dbg("PRISTINE>>\n".$$pristine_body);

  my $uris = $pms->get_uri_detail_list();
  #rewrite the url
  while (my($uri, $info) = each %{$uris}) { 
      if(defined ($info->{types}->{a})){
	  $$pristine_body=~s/$uri/$pms->{PHISHTAG}->{uri}/mg;
      }
  }
  dbg("PRISTINE>>\n".$$pristine_body);
}

1;
__END__

